cimport cython
from cpython.exc cimport PyErr_SetNone

_TAG_TO_NAME = {
    0x0005: "- doctype",
    0x0002: "-text",
    0x0004: "-comment",
}
ctypedef fused str_or_LexborNode:
    str
    bytes
    LexborNode

cdef inline bytes to_bytes(str_or_LexborNode value):
    cdef bytes bytes_val
    if isinstance(value, unicode):
        bytes_val = <bytes>value.encode("utf-8")
    elif isinstance(value, bytes):
        bytes_val = <bytes>value
    return bytes_val


@cython.final
cdef class LexborNode:
    """A class that represents HTML node (element)."""

    @staticmethod
    cdef LexborNode new(lxb_dom_node_t *node, LexborHTMLParser parser):
        cdef LexborNode lxbnode = LexborNode.__new__(LexborNode)
        lxbnode.node = node
        lxbnode.parser = parser
        return lxbnode

    @property
    def mem_id(self):
        return <size_t> self.node

    @property
    def child(self):
        """Alias for the `first_child` property."""
        return self.first_child

    @property
    def first_child(self):
        """Return the first child node."""
        cdef LexborNode node
        if self.node.first_child:
            node = LexborNode.new(<lxb_dom_node_t *> self.node.first_child, self.parser)
            return node
        return None

    @property
    def parent(self):
        """Return the parent node."""
        cdef LexborNode node
        if self.node.parent != NULL:
            node = LexborNode.new(<lxb_dom_node_t *> self.node.parent, self.parser)
            return node
        return None

    @property
    def next(self):
        """Return next node."""
        cdef LexborNode node
        if self.node.next != NULL:
            node = LexborNode.new(<lxb_dom_node_t *> self.node.next, self.parser)
            return node
        return None

    @property
    def prev(self):
        """Return previous node."""
        cdef LexborNode node
        if self.node.prev != NULL:
            node = LexborNode.new(<lxb_dom_node_t *> self.node.prev, self.parser)
            return node
        return None

    @property
    def last_child(self):
        """Return last child node."""
        cdef LexborNode node
        if self.node.last_child != NULL:
            node = LexborNode.new(<lxb_dom_node_t *> self.node.last_child, self.parser)
            return node
        return None

    @property
    def html(self):
        """Return HTML representation of the current node including all its child nodes.

        Returns
        -------
        text : str
        """
        cdef lexbor_str_t *lxb_str
        cdef lxb_status_t status

        lxb_str = lexbor_str_create()
        status = lxb_html_serialize_tree_str(self.node, lxb_str)
        if status == 0 and lxb_str.data:
            html = lxb_str.data.decode(_ENCODING).replace('<-undef>', '')
            lexbor_str_destroy(lxb_str,  self.node.owner_document.text, True)
            return html
        return None

    def __hash__(self):
        return self.mem_id

    def text_lexbor(self):
        """Returns the text of the node including text of all its child nodes.

        Uses builtin method from lexbor.
        """

        cdef size_t str_len = 0
        cdef lxb_char_t * text

        text = lxb_dom_node_text_content(self.node, &str_len)
        if <int>str_len == 0:
            raise RuntimeError("Can't extract text")

        unicode_text = text.decode(_ENCODING)
        lxb_dom_document_destroy_text_noi(self.node.owner_document, text)
        return unicode_text

    def text(self, bool deep=True, str separator='', bool strip=False):
        """Returns the text of the node including text of all its child nodes.

        Parameters
        ----------
        strip : bool, default False
            If true, calls ``str.strip()`` on each text part to remove extra white spaces.
        separator : str, default ''
            The separator to use when joining text from different nodes.
        deep : bool, default True
            If True, includes text from all child nodes.

        Returns
        -------
        text : str

        """
        cdef unsigned char * text
        cdef lxb_dom_node_t* node = <lxb_dom_node_t*> self.node.first_child

        if not deep:
            container = TextContainer(separator, strip)
            if self.node != NULL and self.node.type == LXB_DOM_NODE_TYPE_TEXT:
                text = <unsigned char *> lexbor_str_data_noi(&(<lxb_dom_character_data_t *> self.node).data)
                if text != NULL:
                    py_text = text.decode(_ENCODING)
                    container.append(py_text)

            while node != NULL:
                if node.type == LXB_DOM_NODE_TYPE_TEXT:
                    text = <unsigned char *> lexbor_str_data_noi(&(<lxb_dom_character_data_t *> node).data)
                    if text != NULL:
                        py_text = text.decode(_ENCODING)
                        container.append(py_text)
                node = node.next
            return container.text
        else:
            container = TextContainer(separator, strip)
            if self.node.type == LXB_DOM_NODE_TYPE_TEXT:
                text = <unsigned char *> lexbor_str_data_noi(&(<lxb_dom_character_data_t *> self.node).data)
                if text != NULL:
                    container.append(text.decode(_ENCODING))

            lxb_dom_node_simple_walk(
                <lxb_dom_node_t *> self.node,
                <lxb_dom_node_simple_walker_f>text_callback,
                <void *>container
            )
            return container.text

    def css(self, str query):
        """Evaluate CSS selector against current node and its child nodes.

        Matches pattern `query` against HTML tree.
        `CSS selectors reference <https://www.w3schools.com/cssref/css_selectors.asp>`_.

        Parameters
        ----------
        query : str
            CSS selector (e.g. "div > :nth-child(2n+1):not(:has(a))").

        Returns
        -------
        selector : list of `Node` objects
        """
        return self.parser.selector.find(query, self)

    def css_first(self, str query, default=None, bool strict=False):
        """Same as `css` but returns only the first match.

        Parameters
        ----------

        query : str
        default : bool, default None
            Default value to return if there is no match.
        strict: bool, default True
            Set to True if you want to check if there is strictly only one match in the document.


        Returns
        -------
        selector : `LexborNode` object
        """
        # TODO: This can be improved.
        results = self.css(query)
        n_results = len(results)
        if n_results > 0:
            if strict and n_results > 1:
                raise ValueError("Expected 1 match, but found %s matches" % n_results)
            return results[0]
        return default

    def any_css_matches(self, tuple selectors):
        """Returns True if any of CSS selectors matches a node"""
        for selector in selectors:
            if self.parser.selector.any_matches(selector,  self):
                return True
        return False

    def css_matches(self, str selector):
        """Returns True if CSS selector matches a node."""
        return self.parser.selector.any_matches(selector, self)

    def __repr__(self):
        return '<LexborNode %s>' % self.tag

    @property
    def tag_id(self):
        cdef lxb_tag_id_t tag_id = lxb_dom_node_tag_id_noi(self.node)
        return tag_id

    @property
    def tag(self):
        """Return the name of the current tag (e.g. div, p, img).

        Returns
        -------
        text : str
        """

        cdef lxb_char_t *c_text
        cdef size_t str_len = 0
        if self.tag_id in [LXB_TAG__EM_DOCTYPE, LXB_TAG__TEXT, LXB_TAG__EM_COMMENT]:
            return _TAG_TO_NAME[self.tag_id]
        c_text = lxb_dom_element_qualified_name(<lxb_dom_element_t *> self.node, &str_len)
        text = None
        if c_text:
            text = c_text.decode(_ENCODING)
        return text

    def decompose(self, bool recursive=True):
        """Remove the current node from the tree.

        Parameters
        ----------
        recursive : bool, default True
            Whenever to delete all its child nodes

        Examples
        --------

        >>> tree = LexborHTMLParser(html)
        >>> for tag in tree.css('script'):
        >>>     tag.decompose()

        """
        if self.node == <lxb_dom_node_t *> lxb_dom_document_root(&self.parser.document.dom_document):
            raise SelectolaxError("Decomposing the root node is not allowed.")

        if recursive:
            lxb_dom_node_destroy_deep(<lxb_dom_node_t *> self.node)
        else:
            lxb_dom_node_destroy(<lxb_dom_node_t *> self.node)

    def strip_tags(self, list tags, bool recursive = False):
        """Remove specified tags from the HTML tree.

        Parameters
        ----------
        tags : list
            List of tags to remove.
        recursive : bool, default True
            Whenever to delete all its child nodes

        Examples
        --------

        >>> tree = LexborHTMLParser('<html><head></head><body><script></script><div>Hello world!</div></body></html>')
        >>> tags = ['head', 'style', 'script', 'xmp', 'iframe', 'noembed', 'noframes']
        >>> tree.strip_tags(tags)
        >>> tree.html
        '<html><body><div>Hello world!</div></body></html>'

        """
        cdef LexborNode element
        for tag in tags:
            for element in self.css(tag):
                element.decompose(recursive=recursive)

    @property
    def attributes(self):
        """Get all attributes that belong to the current node.

        The value of empty attributes is None.

        Returns
        -------
        attributes : dictionary of all attributes.

        Examples
        --------

        >>> tree = LexborHTMLParser("<div data id='my_id'></div>")
        >>> node = tree.css_first('div')
        >>> node.attributes
        {'data': None, 'id': 'my_id'}
        """
        cdef lxb_dom_attr_t *attr = lxb_dom_element_first_attribute_noi(<lxb_dom_element_t *> self.node)
        cdef size_t str_len = 0
        attributes = dict()

        while attr != NULL:
            key = lxb_dom_attr_local_name_noi(attr, &str_len)
            value = lxb_dom_attr_value_noi(attr, &str_len)

            if value:
                py_value = value.decode(_ENCODING)
            else:
                py_value = None
            attributes[key.decode(_ENCODING)] = py_value

            attr = attr.next
        return attributes

    @property
    def attrs(self):
        """A dict-like object that is similar to the ``attributes`` property, but operates directly on the Node data.

        .. warning:: Use ``attributes`` instead, if you don't want to modify Node attributes.

        Returns
        -------
        attributes : Attributes mapping object.

        Examples
        --------

        >>> tree = LexborHTMLParser("<div id='a'></div>")
        >>> node = tree.css_first('div')
        >>> node.attrs
        <div attributes, 1 items>
        >>> node.attrs['id']
        'a'
        >>> node.attrs['foo'] = 'bar'
        >>> del node.attrs['id']
        >>> node.attributes
        {'foo': 'bar'}
        >>> node.attrs['id'] = 'new_id'
        >>> node.html
        '<div foo="bar" id="new_id"></div>'
        """
        cdef LexborAttributes attributes = LexborAttributes.create(<lxb_dom_node_t *>self.node)
        return attributes

    @property
    def id(self):
        """Get the id attribute of the node.

        Returns None if id does not set.

        Returns
        -------
        text : str
        """
        cdef char * key = 'id'
        cdef size_t str_len
        cdef lxb_dom_attr_t * attr = lxb_dom_element_attr_by_name(
            <lxb_dom_element_t *> self.node,
            <lxb_char_t *> key, 2
        )
        if attr != NULL:
            value = lxb_dom_attr_value_noi(attr, &str_len)
            return value.decode(_ENCODING) if value else None
        return None

    def iter(self, include_text=False):
        """Iterate over nodes on the current level.

        Parameters
        ----------
        include_text : bool
            If True, includes text nodes as well.

        Yields
        -------
        node
        """

        cdef lxb_dom_node_t *node = self.node.first_child
        cdef LexborNode next_node

        while node != NULL:
            if node.type == LXB_DOM_NODE_TYPE_TEXT and not include_text:
                node = node.next
                continue

            next_node = LexborNode.new(<lxb_dom_node_t *> node, self.parser)
            yield next_node
            node = node.next

    def unwrap(self, bint delete_empty=False):
        """Replace node with whatever is inside this node.

        Parameters
        ----------
        delete_empty : bool, default False
            If True, removes empty tags.

        Examples
        --------

        >>>  tree = LexborHTMLParser("<div>Hello <i>world</i>!</div>")
        >>>  tree.css_first('i').unwrap()
        >>>  tree.html
        '<html><head></head><body><div>Hello world!</div></body></html>'

        Note: by default, empty tags are ignored, use "delete_empty" to change this.
        """
        if self.node.first_child == NULL:
            if delete_empty:
                lxb_dom_node_destroy(<lxb_dom_node_t *> self.node)
            return
        cdef lxb_dom_node_t* next_node
        cdef lxb_dom_node_t* current_node

        if self.node.first_child.next != NULL:
            current_node = self.node.first_child
            next_node = current_node.next

            while next_node != NULL:
                next_node = current_node.next
                lxb_dom_node_insert_before(self.node, current_node)
                current_node = next_node
        else:
            lxb_dom_node_insert_before(self.node, self.node.first_child)
        lxb_dom_node_destroy(<lxb_dom_node_t *> self.node)

    def unwrap_tags(self, list tags, bint delete_empty = False):
        """Unwraps specified tags from the HTML tree.

        Works the same as the ``unwrap`` method, but applied to a list of tags.

        Parameters
        ----------
        tags : list
            List of tags to remove.
        delete_empty : bool, default False
            If True, removes empty tags.

        Examples
        --------

        >>> tree = LexborHTMLParser("<div><a href="">Hello</a> <i>world</i>!</div>")
        >>> tree.body.unwrap_tags(['i','a'])
        >>> tree.body.html
        '<body><div>Hello world!</div></body>'

        Note: by default, empty tags are ignored, use "delete_empty" to change this.
        """
        cdef LexborNode element
        for tag in tags:
            for element in self.css(tag):
                element.unwrap(delete_empty)

    def merge_text_nodes(self):
        """Iterates over all text nodes and merges all text nodes that are close to each other.

        This is useful for text extraction.
        Use it when you need to strip HTML tags and merge "dangling" text.

        Examples
        --------

        >>> tree = LexborHTMLParser("<div><p><strong>J</strong>ohn</p><p>Doe</p></div>")
        >>> node = tree.css_first('div')
        >>> tree.unwrap_tags(["strong"])
        >>> tree.text(deep=True, separator=" ", strip=True)
        "J ohn Doe" # Text extraction produces an extra space because the strong tag was removed.
        >>> node.merge_text_nodes()
        >>> tree.text(deep=True, separator=" ", strip=True)
        "John Doe"
        """
        cdef lxb_dom_node_t *node = self.node.first_child
        cdef lxb_dom_node_t *next_node
        cdef lxb_char_t *left_text
        cdef lxb_char_t *right_text
        cdef size_t left_length, right_length

        while node != NULL:
            next_node = node.next
            if node.type == LXB_DOM_NODE_TYPE_TEXT and node.prev and node.prev.type == LXB_DOM_NODE_TYPE_TEXT:
                left_text = lxb_dom_node_text_content(node.prev, &left_length)
                right_text = lxb_dom_node_text_content(node, &right_length)
                if left_text and right_text:
                    combined = (<bytes>left_text[:left_length]) + (<bytes>right_text[:right_length])
                    lxb_dom_node_text_content_set(node, combined, len(combined))
                    lxb_dom_node_remove(node.prev)
            if node.first_child:
                LexborNode.new(node, self.parser).merge_text_nodes()
            node = next_node

    def traverse(self, include_text=False):
        """Iterate over all child and next nodes starting from the current level.

        Parameters
        ----------
        include_text : bool
            If True, includes text nodes as well.

        Yields
        -------
        node
        """
        cdef lxb_dom_node_t * root = self.node
        cdef lxb_dom_node_t * node = root
        cdef LexborNode lxb_node

        while node != NULL:
            if not (not include_text and node.type == LXB_DOM_NODE_TYPE_TEXT):
                lxb_node = LexborNode.new(<lxb_dom_node_t *> node, self.parser)
                yield lxb_node

            if node.first_child != NULL:
                node = node.first_child
            else:
                while node != root and node.next == NULL:
                    node = node.parent
                if node == root:
                    break
                node = node.next

    def replace_with(self, str_or_LexborNode value):
        """Replace current Node with specified value.

        Parameters
        ----------
        value : str, bytes or Node
            The text or Node instance to replace the Node with.
            When a text string is passed, it's treated as text. All HTML tags will be escaped.
            Convert and pass the ``Node`` object when you want to work with HTML.
            Does not clone the ``Node`` object.
            All future changes to the passed ``Node`` object will also be taken into account.

        Examples
        --------

        >>> tree = LexborHTMLParser('<div>Get <img src="" alt="Laptop"></div>')
        >>> img = tree.css_first('img')
        >>> img.replace_with(img.attributes.get('alt', ''))
        >>> tree.body.child.html
        '<div>Get Laptop</div>'

        >>> html_parser = LexborHTMLParser('<div>Get <span alt="Laptop"><img src="/jpg"> <div></div></span></div>')
        >>> html_parser2 = LexborHTMLParser('<div>Test</div>')
        >>> img_node = html_parser.css_first('img')
        >>> img_node.replace_with(html_parser2.body.child)
        '<div>Get <span alt="Laptop"><div>Test</div> <div></div></span></div>'
        """
        cdef lxb_dom_node_t * new_node

        if isinstance(value, (str, bytes, unicode)):
            bytes_val = to_bytes(value)
            new_node = <lxb_dom_node_t *> lxb_dom_document_create_text_node(
                    &self.parser.document.dom_document,
                    <lxb_char_t *> bytes_val, len(bytes_val)
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_before(self.node,  new_node)
            lxb_dom_node_destroy(<lxb_dom_node_t *> self.node)
        elif isinstance(value, LexborNode):
            new_node = lxb_dom_document_import_node(
                &self.parser.document.dom_document,
                <lxb_dom_node_t *> value.node,
                <bint> True
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_before(self.node, <lxb_dom_node_t *> new_node)
            lxb_dom_node_destroy(<lxb_dom_node_t *> self.node)
        else:
            raise SelectolaxError("Expected a string or LexborNode instance, but %s found" % type(value).__name__)

    def insert_before(self, str_or_LexborNode value):
        """
        Insert a node before the current Node.

        Parameters
        ----------
        value : str, bytes or Node
            The text or Node instance to insert before the Node.
            When a text string is passed, it's treated as text. All HTML tags will be escaped.
            Convert and pass the ``Node`` object when you want to work with HTML.
            Does not clone the ``Node`` object.
            All future changes to the passed ``Node`` object will also be taken into account.

        Examples
        --------

        >>> tree = LexborHTMLParser('<div>Get <img src="" alt="Laptop"></div>')
        >>> img = tree.css_first('img')
        >>> img.insert_before(img.attributes.get('alt', ''))
        >>> tree.body.child.html
        '<div>Get Laptop<img src="" alt="Laptop"></div>'

        >>> html_parser = LexborHTMLParser('<div>Get <span alt="Laptop"><img src="/jpg"> <div></div></span></div>')
        >>> html_parser2 = LexborHTMLParser('<div>Test</div>')
        >>> img_node = html_parser.css_first('img')
        >>> img_node.insert_before(html_parser2.body.child)
        <div>Get <span alt="Laptop"><div>Test</div><img src="/jpg"> <div></div></span></div>'
        """
        cdef lxb_dom_node_t * new_node

        if isinstance(value, (str, bytes, unicode)):
            bytes_val = to_bytes(value)
            new_node = <lxb_dom_node_t *> lxb_dom_document_create_text_node(
                    &self.parser.document.dom_document,
                    <lxb_char_t *> bytes_val, len(bytes_val)
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_before(self.node,  new_node)
        elif isinstance(value, LexborNode):
            new_node = lxb_dom_document_import_node(
                &self.parser.document.dom_document,
                <lxb_dom_node_t *> value.node,
                <bint> True
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_before(self.node, <lxb_dom_node_t *> new_node)
        else:
            raise SelectolaxError("Expected a string or LexborNode instance, but %s found" % type(value).__name__)

    def insert_after(self, str_or_LexborNode value):
        """
        Insert a node after the current Node.

        Parameters
        ----------
        value : str, bytes or Node
            The text or Node instance to insert after the Node.
            When a text string is passed, it's treated as text. All HTML tags will be escaped.
            Convert and pass the ``Node`` object when you want to work with HTML.
            Does not clone the ``Node`` object.
            All future changes to the passed ``Node`` object will also be taken into account.

        Examples
        --------

        >>> tree = LexborHTMLParser('<div>Get <img src="" alt="Laptop"></div>')
        >>> img = tree.css_first('img')
        >>> img.insert_after(img.attributes.get('alt', ''))
        >>> tree.body.child.html
        '<div>Get <img src="" alt="Laptop">Laptop</div>'

        >>> html_parser = LexborHTMLParser('<div>Get <span alt="Laptop"><img src="/jpg"> <div></div></span></div>')
        >>> html_parser2 = LexborHTMLParser('<div>Test</div>')
        >>> img_node = html_parser.css_first('img')
        >>> img_node.insert_after(html_parser2.body.child)
        <div>Get <span alt="Laptop"><img src="/jpg"><div>Test</div> <div></div></span></div>'
        """
        cdef lxb_dom_node_t * new_node

        if isinstance(value, (str, bytes, unicode)):
            bytes_val = to_bytes(value)
            new_node = <lxb_dom_node_t *> lxb_dom_document_create_text_node(
                    &self.parser.document.dom_document,
                    <lxb_char_t *> bytes_val, len(bytes_val)
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_after(self.node,  new_node)
        elif isinstance(value, LexborNode):
            new_node = lxb_dom_document_import_node(
                &self.parser.document.dom_document,
                <lxb_dom_node_t *> value.node,
                <bint> True
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_after(self.node, <lxb_dom_node_t *> new_node)
        else:
            raise SelectolaxError("Expected a string or LexborNode instance, but %s found" % type(value).__name__)

    def insert_child(self, str_or_LexborNode value):
        """
        Insert a node inside (at the end of) the current Node.

        Parameters
        ----------
        value : str, bytes or Node
            The text or Node instance to insert inside the Node.
            When a text string is passed, it's treated as text. All HTML tags will be escaped.
            Convert and pass the ``Node`` object when you want to work with HTML.
            Does not clone the ``Node`` object.
            All future changes to the passed ``Node`` object will also be taken into account.

        Examples
        --------

        >>> tree = LexborHTMLParser('<div>Get <img src=""></div>')
        >>> div = tree.css_first('div')
        >>> div.insert_child('Laptop')
        >>> tree.body.child.html
        '<div>Get <img src="">Laptop</div>'

        >>> html_parser = LexborHTMLParser('<div>Get <span alt="Laptop"> <div>Laptop</div> </span></div>')
        >>> html_parser2 = LexborHTMLParser('<div>Test</div>')
        >>> span_node = html_parser.css_first('span')
        >>> span_node.insert_child(html_parser2.body.child)
        <div>Get <span alt="Laptop"> <div>Laptop</div> <div>Test</div> </span></div>'
        """
        cdef lxb_dom_node_t * new_node

        if isinstance(value, (str, bytes, unicode)):
            bytes_val = to_bytes(value)
            new_node = <lxb_dom_node_t *> lxb_dom_document_create_text_node(
                    &self.parser.document.dom_document,
                    <lxb_char_t *> bytes_val, len(bytes_val)
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_child(self.node,  new_node)
        elif isinstance(value, LexborNode):
            new_node = lxb_dom_document_import_node(
                &self.parser.document.dom_document,
                <lxb_dom_node_t *> value.node,
                <bint> True
            )
            if new_node == NULL:
                raise SelectolaxError("Can't create a new node")
            lxb_dom_node_insert_child(self.node, <lxb_dom_node_t *> new_node)
        else:
            raise SelectolaxError("Expected a string or LexborNode instance, but %s found" % type(value).__name__)

    @property
    def raw_value(self):
        """Return the raw (unparsed, original) value of a node.

        Currently, works on text nodes only.

        Returns
        -------

        raw_value : bytes

        Examples
        --------

        >>> html_parser = LexborHTMLParser('<div>&#x3C;test&#x3E;</div>')
        >>> selector = html_parser.css_first('div')
        >>> selector.child.html
        '&lt;test&gt;'
        >>> selector.child.raw_value
        b'&#x3C;test&#x3E;'
        """
        raise NotImplementedError("This features is not supported by the lexbor backend. Please use Modest backend.")

    def scripts_contain(self, str query):
        """Returns True if any of the script tags contain specified text.

        Caches script tags on the first call to improve performance.

        Parameters
        ----------
        query : str
            The query to check.

        """
        cdef LexborNode node
        if self.parser.cached_script_texts is None:
            nodes = self.parser.selector.find('script', self)
            text_nodes = []
            for node in nodes:
                node_text = node.text(deep=True)
                if node_text:
                    text_nodes.append(node_text)
            self.parser.cached_script_texts = text_nodes

        for text in self.parser.cached_script_texts:
            if query in text:
                return True
        return False

    def script_srcs_contain(self, tuple queries):
        """Returns True if any of the script SRCs attributes contain on of the specified text.

        Caches values on the first call to improve performance.

        Parameters
        ----------
        queries : tuple of str

        """
        cdef LexborNode node
        if self.parser.cached_script_srcs is None:
            nodes = self.parser.selector.find('script', self)
            src_nodes = []
            for node in nodes:
                node_src = node.attrs.get('src')
                if node_src:
                    src_nodes.append(node_src)
            self.parser.cached_script_srcs = src_nodes

        for text in self.parser.cached_script_srcs:
            for query in queries:
                if query in text:
                    return True
        return False

    def remove(self, bool recursive=True):
        """An alias for the decompose method."""
        self.decompose(recursive)

    def select(self, query=None):
        """Select nodes given a CSS selector.

        Works similarly to the the ``css`` method, but supports chained filtering and extra features.

        Parameters
        ----------
        query : str or None
            The CSS selector to use when searching for nodes.

        Returns
        -------
        selector : The `Selector` class.
        """
        return LexborSelector(self, query)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.html == other
        if not isinstance(other, LexborNode):
            return False
        return self.html == other.html

    @property
    def text_content(self):
        """Returns the text of the node if it is a text node.

        Returns None for other nodes.
        Unlike the ``text`` method, does not include child nodes.

        Returns
        -------
        text : str or None.
        """
        cdef unsigned char * text
        cdef lxb_dom_node_t* node = <lxb_dom_node_t*> self.node.first_child
        cdef TextContainer container
        if self.node == NULL or self.node.type != LXB_DOM_NODE_TYPE_TEXT:
            return None

        text = <unsigned char *> lexbor_str_data_noi(&(<lxb_dom_character_data_t *> self.node).data)
        if text != NULL:
            container = TextContainer.new_with_defaults()
            py_text = text.decode(_ENCODING)
            container.append(py_text)
            return container.text


@cython.internal
@cython.final
cdef class TextContainer:
    cdef str _text
    cdef str separator
    cdef bint strip

    @staticmethod
    cdef TextContainer new_with_defaults():
        cdef TextContainer cls = TextContainer.__new__(TextContainer)
        cls._text = ''
        cls.separator = ''
        cls.strip = False
        return cls

    def __init__(self, str separator = '', bool strip = False):
        self._text = ""
        self.separator = separator
        self.strip = strip

    def append(self, str node_text):
        if self.strip:
            self._text += node_text.strip() + self.separator
        else:
            self._text += node_text + self.separator

    @property
    def text(self):
        if self.separator and self._text and self._text.endswith(self.separator):
            self._text = self._text[:-len(self.separator)]
        return self._text


cdef lexbor_action_t text_callback(lxb_dom_node_t *node, void *ctx):
    cdef unsigned char *text
    cdef lxb_tag_id_t tag_id = lxb_dom_node_tag_id_noi(node)
    if tag_id != LXB_TAG__TEXT:
        return LEXBOR_ACTION_OK

    text = <unsigned char*> lexbor_str_data_noi(&(<lxb_dom_text_t *> node).char_data.data)
    if not text:
        return LEXBOR_ACTION_OK

    try:
        py_str = text.decode(_ENCODING)

    except Exception as e:
        PyErr_SetNone(e)
        return LEXBOR_ACTION_STOP

    cdef TextContainer cls
    cls = <TextContainer> ctx
    cls.append(py_str)
    return LEXBOR_ACTION_OK



import threading
from difflib import SequenceMatcher

import pytest
from selectolax.parser import HTMLParser, Node

from selectolax.lexbor import LexborHTMLParser, LexborNode, SelectolaxError, create_tag

"""
We'are testing only our own code.
Many functionality are already tested in the Modest engine, so there is no reason to test every case.
"""

_PARSERS_PARAMETRIZER = (
    "parser",
    (HTMLParser, LexborHTMLParser),
)


def test_encoding():
    html = "<div><p id=p1><p id=p2><p id=p3><a>link</a><p id=p4><p id=p5>text<p id=p6></div>"
    html = HTMLParser(html)
    assert html.input_encoding == "UTF-8"

    html = b"<div><p id=p1><p id=p2><p id=p3><a>link</a><p id=p4><p id=p5>text<p id=p6></div>"
    html = HTMLParser(html)
    assert html.input_encoding == "UTF-8"

    html = "<div>Привет мир!</div>".encode("cp1251")
    assert HTMLParser(html, detect_encoding=True).input_encoding == "WINDOWS-1251"

    html_utf = '<head><meta charset="WINDOWS-1251"></head>'.encode("utf-8")
    assert (
        HTMLParser(html_utf, detect_encoding=True, use_meta_tags=True).input_encoding
        == "WINDOWS-1251"
    )

    # UTF-16 not ASCII-readable
    html_utf = '<head><meta charset="WINDOWS-1251"></head>'.encode("utf-16le")
    assert (
        HTMLParser(html_utf, detect_encoding=True, use_meta_tags=True).input_encoding
        == "UTF-16LE"
    )

    # Unencodable characters in string, should not throw an exception by default
    html_unencodable = b"<div>Roboto+Condensed</div>".decode("utf-7", errors="ignore")
    assert HTMLParser(html_unencodable).input_encoding == "UTF-8"

    # decode_errrors='strict' should error out
    try:
        HTMLParser(html_unencodable, decode_errors="strict")
        assert False
    except Exception as e:
        assert type(e) is UnicodeEncodeError


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_parser(parser):
    html = parser("")
    assert isinstance(html, parser)

    with pytest.raises(TypeError):
        parser(123)

    with pytest.raises(TypeError):
        parser("asd").css(123)


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_malformed_data(parser):
    malformed_inputs = [
        b"\x00\x01\x02\x03",
        "<div><p><span></div>",
        "<" + "a" * 1000 + ">",
    ]

    for malformed_html in malformed_inputs:
        try:
            html_parser = parser(malformed_html)
            # Should not crash, but may return None or empty results
            result = html_parser.html
            assert result is None or isinstance(result, str)
        except (ValueError, RuntimeError, UnicodeDecodeError):
            # These exceptions are acceptable for malformed input
            pass


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_properties(parser):
    html_parser = parser("<div><p>test</p></div>")

    properties_to_test = ["root", "head", "body", "html"]

    for prop_name in properties_to_test:
        getattr(html_parser, prop_name)


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_unicode_handling(parser):
    unicode_content = [
        "Hello 世界",
        "🚀🌟💫",
        "Café résumé naïve",
    ]

    for content in unicode_content:
        html = f"<div>{content}</div>"
        try:
            html_parser = parser(html)
            result = html_parser.css_first("div")
            if result:
                extracted_text = result.text()
                assert content in extracted_text
        except UnicodeEncodeError:
            # Some encoding issues might be expected
            pass


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_tag_name_validation(parser):
    """Test that tag name validation works correctly."""
    html_parser = parser("<div></div>")

    # Empty tag name should be rejected
    with pytest.raises(ValueError, match="Tag name cannot be empty"):
        html_parser.tags("")

    # Very long tag names should be rejected
    long_tag_name = "a" * 101  # Exceeds 100 character limit
    with pytest.raises(ValueError, match="Tag name is too long"):
        html_parser.tags(long_tag_name)


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_nodes(parser):
    html = (
        '<div><p id="p1"></p><p id="p2"></p><p id="p3"><a>link</a></p>'
        '<p id="p4"></p><p id="p5">text</p><p id="p6"></p></div>'
    )
    htmlp = parser(html)

    assert isinstance(htmlp.root, (Node, LexborNode))
    assert isinstance(htmlp.body, (Node, LexborNode))
    html_output = htmlp.html
    assert len(html_output) >= len(html)
    assert SequenceMatcher(None, html, html_output).ratio() > 0.8


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_root_css(parser):
    tree = parser("test")
    assert len(tree.root.css("data")) == 0


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_strip_tags_from_root(parser):
    html = "<body><div></div><script></script></body>"
    html_parser = parser(html)
    html_parser.root.strip_tags(["div", "script"])
    assert html_parser.html == "<html><head></head><body></body></html>"

    with pytest.raises(TypeError):
        html_parser.strip_tags(1)


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_clone(parser):
    html_parser = parser("""<h1>Welcome</h1>""")
    clone = html_parser.clone()
    html_parser.root.css_first("h1").decompose()
    del html_parser
    assert clone.html == "<html><head></head><body><h1>Welcome</h1></body></html>"


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_tags(parser):
    html_parser = parser("""
    <div><span><span></span></span></div>
    <div><span></span></div>
    <div><div></div></div>
    <span></span>
    <div></div>
    """)
    assert len(html_parser.tags("div")) == 5


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_preserves_doctype(parser):
    html_parser = parser("""
    <!DOCTYPE html>
    <html>
        <head><title>Test</title></head>
        <body><p>Hello World</p></body>
    </html>
    """)
    assert "<!DOCTYPE html>" in html_parser.html


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_invalid_input_types(parser):
    with pytest.raises(TypeError, match="Expected a string"):
        parser(123)

    with pytest.raises(TypeError, match="Expected a string"):
        parser([])

    with pytest.raises(TypeError, match="Expected a string"):
        parser(None)


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_clone_handling(parser):
    html_parser = parser("<div>test</div>")

    cloned = html_parser.clone()
    assert cloned.html is not None

    assert html_parser.html is not None


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_concurrent_parsing(parser):
    """Test that concurrent parsing doesn't cause race conditions."""
    results = []
    errors = []
    lock = threading.Lock()

    def parse_html(content):
        try:
            html_parser = parser(content)
            result = html_parser.body.text()
            if result:
                with lock:
                    results.append(result)
        except Exception as e:
            with lock:
                errors.append(e)

    threads = []
    test_content = "<div>Content {}</div>"

    for i in range(50):
        content = test_content.format(i)
        t1 = threading.Thread(target=parse_html, args=(content,))
        threads.append(t1)

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert len(errors) == 0
    assert len(results) == 50


def test_css_selector_error_handling():
    html_parser = LexborHTMLParser("<div class='test'>content</div>")

    # Invalid selector types should raise TypeError
    with pytest.raises(TypeError):
        html_parser.css(123)

    with pytest.raises(TypeError):
        html_parser.css(None)

    invalid_selectors = [
        ":::",
        "[[[",
        "div{color:red}",
        'h3:contains("some substring")',
    ]

    for selector in invalid_selectors:
        try:
            result = html_parser.css(selector)
            # Should return empty list or raise specific exception
            assert isinstance(result, list)
        except SelectolaxError:
            # Specific parsing errors are acceptable
            pass


@pytest.mark.parametrize(*_PARSERS_PARAMETRIZER)
def test_null_pointer_safety(parser):
    """Test that NULL pointer checks prevent crashes."""
    # Test edge cases that might result in NULL pointers
    edge_cases = [
        "",  # Empty HTML
        "<>",  # Empty tag
        "<!>",  # Empty declaration
        "<html></html>",  # Minimal valid HTML
    ]
    properties_to_test = ["root", "head", "body", "html"]
    for html_content in edge_cases:
        html_parser = parser(html_content)

        for prop_name in properties_to_test:
            getattr(html_parser, prop_name)


def test_decompose_root_node():
    html_parser = LexborHTMLParser("<div><p>test</p></div>")
    with pytest.raises(SelectolaxError):
        html_parser.root.decompose()


def test_empty_attribute_lexbor():
    div = create_tag("div")
    div.attrs["hidden"] = None
    assert div.html == "<div hidden></div>"



cimport cython


@cython.final
cdef class LexborAttributes:
    """A dict-like object that represents attributes."""
    cdef lxb_dom_node_t *node
    cdef unicode decode_errors

    @staticmethod
    cdef LexborAttributes create(lxb_dom_node_t *node):
        obj = <LexborAttributes> LexborAttributes.__new__(LexborAttributes)
        obj.node = node
        return obj

    def __iter__(self):
        cdef lxb_dom_attr_t *attr = lxb_dom_element_first_attribute_noi(<lxb_dom_element_t *> self.node)
        cdef size_t str_len = 0
        attributes = dict()

        while attr != NULL:
            key = lxb_dom_attr_local_name_noi(attr, &str_len)
            if key is not NULL:
                yield key.decode(_ENCODING)
            attr = attr.next

    def __setitem__(self, str key, object value):
        value = value
        bytes_key = key.encode(_ENCODING)
        bytes_value = value.encode(_ENCODING) if value else b""
        cdef lxb_dom_attr_t *attr
        cdef lxb_dom_document_t *doc

        if value is None:
            # N.B. This is suboptimal, but there is not API to set empty attributes
            attr = lxb_dom_element_set_attribute(
                <lxb_dom_element_t *> self.node,
                <lxb_char_t *> bytes_key, len(bytes_key),
                NULL, 0
            )
            doc = (<lxb_dom_node_t*>attr).owner_document
            lexbor_str_destroy(attr.value, doc.text, 0)
            attr.value = NULL

        elif isinstance(value, str) or isinstance(value, unicode) :
            lxb_dom_element_set_attribute(
                <lxb_dom_element_t *> self.node,
                <lxb_char_t *> bytes_key, len(bytes_key),
                <lxb_char_t *> bytes_value, len(bytes_value),
            )
        else:
            raise TypeError("Expected str or unicode, got %s" % type(value))

    def __delitem__(self, key):
        try:
            self.__getitem__(key)
        except KeyError:
            raise KeyError(key)
        bytes_key = key.encode(_ENCODING)
        lxb_dom_element_remove_attribute(
            <lxb_dom_element_t *> self.node,
            <lxb_char_t *> bytes_key, len(bytes_key),
        )

    def __getitem__(self, str key):
        bytes_key = key.encode(_ENCODING)
        cdef lxb_dom_attr_t * attr = lxb_dom_element_attr_by_name(
            <lxb_dom_element_t *> self.node,
            <lxb_char_t *> bytes_key, len(bytes_key)
        )
        cdef size_t str_len = 0
        if attr != NULL:
            value = lxb_dom_attr_value_noi(attr, &str_len)
            return value.decode(_ENCODING) if value else None
        raise KeyError(key)

    def __len__(self):
        return len(list(self.__iter__()))

    def keys(self):
        return self.__iter__()

    def items(self):
        for key in self.__iter__():
            yield key, self[key]

    def values(self):
        for key in self.__iter__():
            yield self[key]

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def sget(self, key, default=""):
        """Same as get, but returns empty strings instead of None values for empty attributes."""
        try:
            val = self[key]
            if val is None:
                val = ""
            return val
        except KeyError:
            return default

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def __repr__(self):
        cdef lxb_char_t *c_text
        cdef size_t str_len = 0
        c_text = lxb_dom_element_qualified_name(<lxb_dom_element_t *> self.node, &str_len)
        tag_name = c_text.decode(_ENCODING, 'ignore') if c_text != NULL else 'unknown'
        return "<%s attributes, %s items>" % (tag_name, len(self))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import platform

import sys
from setuptools import setup, find_packages, Extension

with io.open("README.rst", mode="rt", encoding="utf-8") as readme_file:
    readme = readme_file.read()

# Setup flags
USE_STATIC = False
USE_CYTHON = False
PLATFORM = "windows_nt" if platform.system() == "Windows" else "posix"
INCLUDE_LEXBOR = bool(os.environ.get("USE_LEXBOR", True))
ARCH = platform.architecture()[0]

try:
    from Cython.Build import cythonize

    HAS_CYTHON = True
    USE_CYTHON = True
except ImportError as err:
    HAS_CYTHON = False

if "--static" in sys.argv:
    USE_STATIC = True
    sys.argv.remove("--static")

if "--lexbor" in sys.argv:
    INCLUDE_LEXBOR = True
    sys.argv.remove("--lexbor")

if "--cython" in sys.argv:
    if HAS_CYTHON:
        USE_CYTHON = True
    else:
        raise ImportError("No module named 'Cython'")
    sys.argv.remove("--cython")

# If there are no pretranspiled source files
if HAS_CYTHON and not os.path.exists("selectolax/parser.c"):
    USE_CYTHON = True

COMPILER_DIRECTIVES = {
    "language_level": 3,
    "embedsignature": True,
    "annotation_typing": False,
    "emit_code_comments": True,
    "boundscheck": False,
    "wraparound": False,
}


def find_modest_files(modest_path="modest/source"):
    c_files = []
    if os.path.exists(modest_path):
        for root, dirs, files in os.walk(modest_path):
            for file in files:
                if file.endswith(".c"):
                    file_path = os.path.join(root, file)

                    # Filter platform specific files
                    if (file_path.find("myport") >= 0) and (
                        not file_path.find(PLATFORM) >= 0
                    ):
                        continue

                    if INCLUDE_LEXBOR:
                        if (file_path.find("ports") >= 0) and (
                            not file_path.find(PLATFORM) >= 0
                        ):
                            continue
                    c_files.append(file_path)
    return c_files


def make_extensions():
    files_to_compile_lxb = []
    extra_objects_lxb, extra_objects = [], []

    if USE_CYTHON:
        files_to_compile = [
            "selectolax/parser.pyx",
        ]
        if INCLUDE_LEXBOR:
            files_to_compile_lxb = [
                "selectolax/lexbor.pyx",
            ]
    else:
        files_to_compile = ["selectolax/parser.c"]
        if INCLUDE_LEXBOR:
            files_to_compile_lxb = [
                "selectolax/lexbor.c",
            ]

    if USE_STATIC:
        extra_objects = ["modest/lib/libmodest_static.a"]
        if INCLUDE_LEXBOR:
            extra_objects_lxb = ["lexbor/liblexbor_static.a"]
    else:
        files_to_compile.extend(find_modest_files("modest/source"))
        if INCLUDE_LEXBOR:
            files_to_compile_lxb.extend(find_modest_files("lexbor/source"))

    compile_arguments_lxb = [
        "-DLEXBOR_STATIC",
    ]
    compile_arguments = [
        "-DMODEST_BUILD_OS=%s" % platform.system(),
        "-DMyCORE_OS_%s" % platform.system(),
        "-DMODEST_PORT_NAME=%s" % PLATFORM,
        "-DMyCORE_BUILD_WITHOUT_THREADS=YES",
        "-DMyCORE_BUILD_DEBUG=NO",
    ]

    if PLATFORM == "posix":
        args = [
            "-pedantic",
            "-fPIC",
            "-Wno-unused-variable",
            "-Wno-unused-function",
            "-std=c99",
            "-O2",
        ]
        compile_arguments.extend(args)
        compile_arguments_lxb.extend(args)
    elif PLATFORM == "windows_nt":
        compile_arguments_lxb.extend(
            [
                "-D_WIN64" if ARCH == "64bit" else "-D_WIN32",
            ]
        )

    extensions = [
        Extension(
            "selectolax.parser",
            files_to_compile,
            language="c",
            include_dirs=[
                "modest/include/",
            ],
            extra_objects=extra_objects,
            extra_compile_args=compile_arguments,
        ),
    ]
    if INCLUDE_LEXBOR:
        extensions.append(
            Extension(
                "selectolax.lexbor",
                files_to_compile_lxb,
                language="c",
                include_dirs=[
                    "lexbor/source/",
                ],
                extra_objects=extra_objects_lxb,
                extra_compile_args=compile_arguments_lxb,
            )
        )
    if USE_CYTHON:
        extensions = cythonize(extensions, compiler_directives=COMPILER_DIRECTIVES)

    return extensions


setup(
    name="selectolax",
    version="0.3.33",
    description="Fast HTML5 parser with CSS selectors.",
    long_description=readme,
    author="Artem Golubin",
    author_email="me@rushter.com",
    url="https://github.com/rushter/selectolax",
    packages=find_packages(include=["selectolax"]),
    package_data={"selectolax": ["py.typed"]},
    include_package_data=True,
    zip_safe=False,
    ext_modules=make_extensions(),
)



from typing import Literal, Optional, Union, Type

MAX_HTML_INPUT_SIZE = 250e+7

ParserCls = Union[Type["HTMLParser"], Type["LexborHTMLParser"]]
Parser = Union["HTMLParser", "LexborHTMLParser"]
FRAGMENT = Literal[
    "document",
    "fragment",
    "head",
    "body",
    "head_and_body",
    "document_no_head",
    "document_no_body",
    "document_no_head_no_body",
]


def preprocess_input(html, decode_errors='ignore'):
    if isinstance(html, (str, unicode)):
        bytes_html = html.encode('UTF-8', errors=decode_errors)
    elif isinstance(html, bytes):
        bytes_html = html
    else:
        raise TypeError("Expected a string, but %s found" % type(html).__name__)
    html_len = len(bytes_html)
    if html_len > MAX_HTML_INPUT_SIZE:
        raise ValueError("The specified HTML input is too large to be processed (%d bytes)" % html_len)
    return bytes_html, html_len


def do_create_tag(tag: str, parser_cls: ParserCls):
    if not tag:
        raise ValueError("Tag name cannot be empty")
    return do_parse_fragment(f"<{tag}></{tag}>", parser_cls)[0]


def get_fragment_type(
    html: str,
    parser_cls: ParserCls,
    tree: Optional[Parser] = None,
) -> FRAGMENT:
    if not tree:
        tree = parser_cls(html)

    import re
    html_re = re.compile(r"<html|<body|<head(?!er)", re.IGNORECASE)

    has_html = False
    has_head = False
    has_body = False
    for match in html_re.finditer(html):
        if match[0] == "<html":
            has_html = True
        elif match[0] == "<head":
            has_head = True
        elif match[0] == "<body":
            has_body = True

        if has_html and has_head and has_body:
            break

    if has_html and has_head and has_body:
        return "document"
    elif has_html and not has_head and has_body:
        return "document_no_head"
    elif has_html and has_head and not has_body:
        return "document_no_body"
    elif has_html and not has_head and not has_body:
        return "document_no_head_no_body"
    elif has_head and not has_body:
        return "head"
    elif not has_head and has_body:
        return "body"
    elif has_head and has_body:
        return "head_and_body"
    else:
        return "fragment"


def do_parse_fragment(html: str, parser_cls: ParserCls):
    """
    Given HTML, parse it into a list of Nodes, such that the nodes
    correspond to the given HTML.

    For contrast, HTMLParser adds `<html>`, `<head>`, and `<body>` tags
    if they are missing. This function does not add these tags.
    """
    html = html.strip()
    tree = parser_cls(html)
    frag_type = get_fragment_type(html, parser_cls, tree)

    if frag_type == "document":
        return [tree.root]
    if frag_type == "document_no_head":
        tree.head.decompose(recursive=True)
        return [tree.root]
    if frag_type == "document_no_body":
        tree.body.decompose(recursive=True)
        return [tree.root]
    if frag_type == "document_no_head_no_body":
        tree.head.decompose(recursive=True)
        tree.body.decompose(recursive=True)
        return [tree.root]
    elif frag_type == "head":
        tree.body.decompose(recursive=True)
        return [tree.head]
    elif frag_type == "body":
        tree.head.decompose(recursive=True)
        return [tree.body]
    elif frag_type == "head_and_body":
        return [tree.head, tree.body]
    else:
        return [
            *tree.head.iter(include_text=True),
            *tree.body.iter(include_text=True),
        ]




cdef extern from "myhtml/myhtml.h" nogil:
    ctypedef unsigned int mystatus_t
    ctypedef struct myhtml_t
    ctypedef size_t myhtml_tag_id_t

    ctypedef struct myhtml_tree_t:
        # not completed struct
        myhtml_t* myhtml
        myhtml_tree_node_t*   document
        myhtml_tree_node_t*   node_html

    ctypedef struct mchar_async_t
    ctypedef struct mycore_string_t:
        char*  data
        size_t size
        size_t length

        mchar_async_t *mchar
        size_t node_idx

    ctypedef struct mycore_string_raw_t:
        char*  data
        size_t size
        size_t length

        myhtml_namespace ns

    ctypedef enum myhtml_options:
        MyHTML_OPTIONS_DEFAULT                 = 0x00
        MyHTML_OPTIONS_PARSE_MODE_SINGLE       = 0x01
        MyHTML_OPTIONS_PARSE_MODE_ALL_IN_ONE   = 0x02
        MyHTML_OPTIONS_PARSE_MODE_SEPARATELY   = 0x04

    ctypedef struct myhtml_collection_t:
        myhtml_tree_node_t **list
        size_t size
        size_t length

    ctypedef  struct myhtml_tree_node_t:
        myhtml_tree_node_flags flags

        myhtml_tag_id_t tag_id
        myhtml_namespace ns

        myhtml_tree_node_t* prev
        myhtml_tree_node_t* next
        myhtml_tree_node_t* child
        myhtml_tree_node_t* parent

        myhtml_tree_node_t* last_child

        myhtml_token_node_t* token
        void* data

        myhtml_tree_t* tree

    ctypedef enum myhtml_namespace:
        MyHTML_NAMESPACE_UNDEF      = 0x00
        MyHTML_NAMESPACE_HTML       = 0x01
        MyHTML_NAMESPACE_MATHML     = 0x02
        MyHTML_NAMESPACE_SVG        = 0x03
        MyHTML_NAMESPACE_XLINK      = 0x04
        MyHTML_NAMESPACE_XML        = 0x05
        MyHTML_NAMESPACE_XMLNS      = 0x06
        MyHTML_NAMESPACE_ANY        = 0x07
        MyHTML_NAMESPACE_LAST_ENTRY = 0x07

    ctypedef enum myhtml_tree_node_flags:
        MyHTML_TREE_NODE_UNDEF           = 0
        MyHTML_TREE_NODE_PARSER_INSERTED = 1
        MyHTML_TREE_NODE_BLOCKING        = 2

    ctypedef enum myhtml_token_type:
        MyHTML_TOKEN_TYPE_OPEN             = 0x000
        MyHTML_TOKEN_TYPE_CLOSE            = 0x001
        MyHTML_TOKEN_TYPE_CLOSE_SELF       = 0x002
        MyHTML_TOKEN_TYPE_DONE             = 0x004
        MyHTML_TOKEN_TYPE_WHITESPACE       = 0x008
        MyHTML_TOKEN_TYPE_RCDATA           = 0x010
        MyHTML_TOKEN_TYPE_RAWTEXT          = 0x020
        MyHTML_TOKEN_TYPE_SCRIPT           = 0x040
        MyHTML_TOKEN_TYPE_PLAINTEXT        = 0x080
        MyHTML_TOKEN_TYPE_CDATA            = 0x100
        MyHTML_TOKEN_TYPE_DATA             = 0x200
        MyHTML_TOKEN_TYPE_COMMENT          = 0x400
        MyHTML_TOKEN_TYPE_NULL             = 0x800

    ctypedef enum myhtml_tags:
        MyHTML_TAG__UNDEF              = 0x000
        MyHTML_TAG__TEXT               = 0x001
        MyHTML_TAG__COMMENT            = 0x002
        MyHTML_TAG__DOCTYPE            = 0x003
        MyHTML_TAG_A                   = 0x004
        MyHTML_TAG_ABBR                = 0x005
        MyHTML_TAG_ACRONYM             = 0x006
        MyHTML_TAG_ADDRESS             = 0x007
        MyHTML_TAG_ANNOTATION_XML      = 0x008
        MyHTML_TAG_APPLET              = 0x009
        MyHTML_TAG_AREA                = 0x00a
        MyHTML_TAG_ARTICLE             = 0x00b
        MyHTML_TAG_ASIDE               = 0x00c
        MyHTML_TAG_AUDIO               = 0x00d
        MyHTML_TAG_B                   = 0x00e
        MyHTML_TAG_BASE                = 0x00f
        MyHTML_TAG_BASEFONT            = 0x010
        MyHTML_TAG_BDI                 = 0x011
        MyHTML_TAG_BDO                 = 0x012
        MyHTML_TAG_BGSOUND             = 0x013
        MyHTML_TAG_BIG                 = 0x014
        MyHTML_TAG_BLINK               = 0x015
        MyHTML_TAG_BLOCKQUOTE          = 0x016
        MyHTML_TAG_BODY                = 0x017
        MyHTML_TAG_BR                  = 0x018
        MyHTML_TAG_BUTTON              = 0x019
        MyHTML_TAG_CANVAS              = 0x01a
        MyHTML_TAG_CAPTION             = 0x01b
        MyHTML_TAG_CENTER              = 0x01c
        MyHTML_TAG_CITE                = 0x01d
        MyHTML_TAG_CODE                = 0x01e
        MyHTML_TAG_COL                 = 0x01f
        MyHTML_TAG_COLGROUP            = 0x020
        MyHTML_TAG_COMMAND             = 0x021
        MyHTML_TAG_COMMENT             = 0x022
        MyHTML_TAG_DATALIST            = 0x023
        MyHTML_TAG_DD                  = 0x024
        MyHTML_TAG_DEL                 = 0x025
        MyHTML_TAG_DETAILS             = 0x026
        MyHTML_TAG_DFN                 = 0x027
        MyHTML_TAG_DIALOG              = 0x028
        MyHTML_TAG_DIR                 = 0x029
        MyHTML_TAG_DIV                 = 0x02a
        MyHTML_TAG_DL                  = 0x02b
        MyHTML_TAG_DT                  = 0x02c
        MyHTML_TAG_EM                  = 0x02d
        MyHTML_TAG_EMBED               = 0x02e
        MyHTML_TAG_FIELDSET            = 0x02f
        MyHTML_TAG_FIGCAPTION          = 0x030
        MyHTML_TAG_FIGURE              = 0x031
        MyHTML_TAG_FONT                = 0x032
        MyHTML_TAG_FOOTER              = 0x033
        MyHTML_TAG_FORM                = 0x034
        MyHTML_TAG_FRAME               = 0x035
        MyHTML_TAG_FRAMESET            = 0x036
        MyHTML_TAG_H1                  = 0x037
        MyHTML_TAG_H2                  = 0x038
        MyHTML_TAG_H3                  = 0x039
        MyHTML_TAG_H4                  = 0x03a
        MyHTML_TAG_H5                  = 0x03b
        MyHTML_TAG_H6                  = 0x03c
        MyHTML_TAG_HEAD                = 0x03d
        MyHTML_TAG_HEADER              = 0x03e
        MyHTML_TAG_HGROUP              = 0x03f
        MyHTML_TAG_HR                  = 0x040
        MyHTML_TAG_HTML                = 0x041
        MyHTML_TAG_I                   = 0x042
        MyHTML_TAG_IFRAME              = 0x043
        MyHTML_TAG_IMAGE               = 0x044
        MyHTML_TAG_IMG                 = 0x045
        MyHTML_TAG_INPUT               = 0x046
        MyHTML_TAG_INS                 = 0x047
        MyHTML_TAG_ISINDEX             = 0x048
        MyHTML_TAG_KBD                 = 0x049
        MyHTML_TAG_KEYGEN              = 0x04a
        MyHTML_TAG_LABEL               = 0x04b
        MyHTML_TAG_LEGEND              = 0x04c
        MyHTML_TAG_LI                  = 0x04d
        MyHTML_TAG_LINK                = 0x04e
        MyHTML_TAG_LISTING             = 0x04f
        MyHTML_TAG_MAIN                = 0x050
        MyHTML_TAG_MAP                 = 0x051
        MyHTML_TAG_MARK                = 0x052
        MyHTML_TAG_MARQUEE             = 0x053
        MyHTML_TAG_MENU                = 0x054
        MyHTML_TAG_MENUITEM            = 0x055
        MyHTML_TAG_META                = 0x056
        MyHTML_TAG_METER               = 0x057
        MyHTML_TAG_MTEXT               = 0x058
        MyHTML_TAG_NAV                 = 0x059
        MyHTML_TAG_NOBR                = 0x05a
        MyHTML_TAG_NOEMBED             = 0x05b
        MyHTML_TAG_NOFRAMES            = 0x05c
        MyHTML_TAG_NOSCRIPT            = 0x05d
        MyHTML_TAG_OBJECT              = 0x05e
        MyHTML_TAG_OL                  = 0x05f
        MyHTML_TAG_OPTGROUP            = 0x060
        MyHTML_TAG_OPTION              = 0x061
        MyHTML_TAG_OUTPUT              = 0x062
        MyHTML_TAG_P                   = 0x063
        MyHTML_TAG_PARAM               = 0x064
        MyHTML_TAG_PLAINTEXT           = 0x065
        MyHTML_TAG_PRE                 = 0x066
        MyHTML_TAG_PROGRESS            = 0x067
        MyHTML_TAG_Q                   = 0x068
        MyHTML_TAG_RB                  = 0x069
        MyHTML_TAG_RP                  = 0x06a
        MyHTML_TAG_RT                  = 0x06b
        MyHTML_TAG_RTC                 = 0x06c
        MyHTML_TAG_RUBY                = 0x06d
        MyHTML_TAG_S                   = 0x06e
        MyHTML_TAG_SAMP                = 0x06f
        MyHTML_TAG_SCRIPT              = 0x070
        MyHTML_TAG_SECTION             = 0x071
        MyHTML_TAG_SELECT              = 0x072
        MyHTML_TAG_SMALL               = 0x073
        MyHTML_TAG_SOURCE              = 0x074
        MyHTML_TAG_SPAN                = 0x075
        MyHTML_TAG_STRIKE              = 0x076
        MyHTML_TAG_STRONG              = 0x077
        MyHTML_TAG_STYLE               = 0x078
        MyHTML_TAG_SUB                 = 0x079
        MyHTML_TAG_SUMMARY             = 0x07a
        MyHTML_TAG_SUP                 = 0x07b
        MyHTML_TAG_SVG                 = 0x07c
        MyHTML_TAG_TABLE               = 0x07d
        MyHTML_TAG_TBODY               = 0x07e
        MyHTML_TAG_TD                  = 0x07f
        MyHTML_TAG_TEMPLATE            = 0x080
        MyHTML_TAG_TEXTAREA            = 0x081
        MyHTML_TAG_TFOOT               = 0x082
        MyHTML_TAG_TH                  = 0x083
        MyHTML_TAG_THEAD               = 0x084
        MyHTML_TAG_TIME                = 0x085
        MyHTML_TAG_TITLE               = 0x086
        MyHTML_TAG_TR                  = 0x087
        MyHTML_TAG_TRACK               = 0x088
        MyHTML_TAG_TT                  = 0x089
        MyHTML_TAG_U                   = 0x08a
        MyHTML_TAG_UL                  = 0x08b
        MyHTML_TAG_VAR                 = 0x08c
        MyHTML_TAG_VIDEO               = 0x08d
        MyHTML_TAG_WBR                 = 0x08e
        MyHTML_TAG_XMP                 = 0x08f
        MyHTML_TAG_ALTGLYPH            = 0x090
        MyHTML_TAG_ALTGLYPHDEF         = 0x091
        MyHTML_TAG_ALTGLYPHITEM        = 0x092
        MyHTML_TAG_ANIMATE             = 0x093
        MyHTML_TAG_ANIMATECOLOR        = 0x094
        MyHTML_TAG_ANIMATEMOTION       = 0x095
        MyHTML_TAG_ANIMATETRANSFORM    = 0x096
        MyHTML_TAG_CIRCLE              = 0x097
        MyHTML_TAG_CLIPPATH            = 0x098
        MyHTML_TAG_COLOR_PROFILE       = 0x099
        MyHTML_TAG_CURSOR              = 0x09a
        MyHTML_TAG_DEFS                = 0x09b
        MyHTML_TAG_DESC                = 0x09c
        MyHTML_TAG_ELLIPSE             = 0x09d
        MyHTML_TAG_FEBLEND             = 0x09e
        MyHTML_TAG_FECOLORMATRIX       = 0x09f
        MyHTML_TAG_FECOMPONENTTRANSFER = 0x0a0
        MyHTML_TAG_FECOMPOSITE         = 0x0a1
        MyHTML_TAG_FECONVOLVEMATRIX    = 0x0a2
        MyHTML_TAG_FEDIFFUSELIGHTING   = 0x0a3
        MyHTML_TAG_FEDISPLACEMENTMAP   = 0x0a4
        MyHTML_TAG_FEDISTANTLIGHT      = 0x0a5
        MyHTML_TAG_FEDROPSHADOW        = 0x0a6
        MyHTML_TAG_FEFLOOD             = 0x0a7
        MyHTML_TAG_FEFUNCA             = 0x0a8
        MyHTML_TAG_FEFUNCB             = 0x0a9
        MyHTML_TAG_FEFUNCG             = 0x0aa
        MyHTML_TAG_FEFUNCR             = 0x0ab
        MyHTML_TAG_FEGAUSSIANBLUR      = 0x0ac
        MyHTML_TAG_FEIMAGE             = 0x0ad
        MyHTML_TAG_FEMERGE             = 0x0ae
        MyHTML_TAG_FEMERGENODE         = 0x0af
        MyHTML_TAG_FEMORPHOLOGY        = 0x0b0
        MyHTML_TAG_FEOFFSET            = 0x0b1
        MyHTML_TAG_FEPOINTLIGHT        = 0x0b2
        MyHTML_TAG_FESPECULARLIGHTING  = 0x0b3
        MyHTML_TAG_FESPOTLIGHT         = 0x0b4
        MyHTML_TAG_FETILE              = 0x0b5
        MyHTML_TAG_FETURBULENCE        = 0x0b6
        MyHTML_TAG_FILTER              = 0x0b7
        MyHTML_TAG_FONT_FACE           = 0x0b8
        MyHTML_TAG_FONT_FACE_FORMAT    = 0x0b9
        MyHTML_TAG_FONT_FACE_NAME      = 0x0ba
        MyHTML_TAG_FONT_FACE_SRC       = 0x0bb
        MyHTML_TAG_FONT_FACE_URI       = 0x0bc
        MyHTML_TAG_FOREIGNOBJECT       = 0x0bd
        MyHTML_TAG_G                   = 0x0be
        MyHTML_TAG_GLYPH               = 0x0bf
        MyHTML_TAG_GLYPHREF            = 0x0c0
        MyHTML_TAG_HKERN               = 0x0c1
        MyHTML_TAG_LINE                = 0x0c2
        MyHTML_TAG_LINEARGRADIENT      = 0x0c3
        MyHTML_TAG_MARKER              = 0x0c4
        MyHTML_TAG_MASK                = 0x0c5
        MyHTML_TAG_METADATA            = 0x0c6
        MyHTML_TAG_MISSING_GLYPH       = 0x0c7
        MyHTML_TAG_MPATH               = 0x0c8
        MyHTML_TAG_PATH                = 0x0c9
        MyHTML_TAG_PATTERN             = 0x0ca
        MyHTML_TAG_POLYGON             = 0x0cb
        MyHTML_TAG_POLYLINE            = 0x0cc
        MyHTML_TAG_RADIALGRADIENT      = 0x0cd
        MyHTML_TAG_RECT                = 0x0ce
        MyHTML_TAG_SET                 = 0x0cf
        MyHTML_TAG_STOP                = 0x0d0
        MyHTML_TAG_SWITCH              = 0x0d1
        MyHTML_TAG_SYMBOL              = 0x0d2
        MyHTML_TAG_TEXT                = 0x0d3
        MyHTML_TAG_TEXTPATH            = 0x0d4
        MyHTML_TAG_TREF                = 0x0d5
        MyHTML_TAG_TSPAN               = 0x0d6
        MyHTML_TAG_USE                 = 0x0d7
        MyHTML_TAG_VIEW                = 0x0d8
        MyHTML_TAG_VKERN               = 0x0d9
        MyHTML_TAG_MATH                = 0x0da
        MyHTML_TAG_MACTION             = 0x0db
        MyHTML_TAG_MALIGNGROUP         = 0x0dc
        MyHTML_TAG_MALIGNMARK          = 0x0dd
        MyHTML_TAG_MENCLOSE            = 0x0de
        MyHTML_TAG_MERROR              = 0x0df
        MyHTML_TAG_MFENCED             = 0x0e0
        MyHTML_TAG_MFRAC               = 0x0e1
        MyHTML_TAG_MGLYPH              = 0x0e2
        MyHTML_TAG_MI                  = 0x0e3
        MyHTML_TAG_MLABELEDTR          = 0x0e4
        MyHTML_TAG_MLONGDIV            = 0x0e5
        MyHTML_TAG_MMULTISCRIPTS       = 0x0e6
        MyHTML_TAG_MN                  = 0x0e7
        MyHTML_TAG_MO                  = 0x0e8
        MyHTML_TAG_MOVER               = 0x0e9
        MyHTML_TAG_MPADDED             = 0x0ea
        MyHTML_TAG_MPHANTOM            = 0x0eb
        MyHTML_TAG_MROOT               = 0x0ec
        MyHTML_TAG_MROW                = 0x0ed
        MyHTML_TAG_MS                  = 0x0ee
        MyHTML_TAG_MSCARRIES           = 0x0ef
        MyHTML_TAG_MSCARRY             = 0x0f0
        MyHTML_TAG_MSGROUP             = 0x0f1
        MyHTML_TAG_MSLINE              = 0x0f2
        MyHTML_TAG_MSPACE              = 0x0f3
        MyHTML_TAG_MSQRT               = 0x0f4
        MyHTML_TAG_MSROW               = 0x0f5
        MyHTML_TAG_MSTACK              = 0x0f6
        MyHTML_TAG_MSTYLE              = 0x0f7
        MyHTML_TAG_MSUB                = 0x0f8
        MyHTML_TAG_MSUP                = 0x0f9
        MyHTML_TAG_MSUBSUP             = 0x0fa
        MyHTML_TAG__END_OF_FILE        = 0x0fb
        MyHTML_TAG_FIRST_ENTRY         = MyHTML_TAG__TEXT
        MyHTML_TAG_LAST_ENTRY          = 0x0fc

    ctypedef enum myhtml_tree_parse_flags_t:
        MyHTML_TREE_PARSE_FLAGS_CLEAN = 0x000
        MyHTML_TREE_PARSE_FLAGS_WITHOUT_BUILD_TREE = 0x001
        MyHTML_TREE_PARSE_FLAGS_WITHOUT_PROCESS_TOKEN = 0x003
        MyHTML_TREE_PARSE_FLAGS_SKIP_WHITESPACE_TOKEN = 0x004
        MyHTML_TREE_PARSE_FLAGS_WITHOUT_DOCTYPE_IN_TREE = 0x008

    ctypedef  struct myhtml_token_node_t:
        myhtml_tag_id_t tag_id

        mycore_string_t str

        size_t raw_begin
        size_t raw_length

        size_t element_begin
        size_t element_length

        myhtml_token_attr_t* attr_first
        myhtml_token_attr_t* attr_last

        myhtml_token_type type

    ctypedef struct myhtml_token_attr_t:
        myhtml_token_attr_t* next
        myhtml_token_attr_t* prev

        mycore_string_t key
        mycore_string_t value

        size_t raw_key_begin
        size_t raw_key_length
        size_t raw_value_begin
        size_t raw_value_length

        myhtml_namespace ns

    ctypedef struct myhtml_tree_attr_t:
        myhtml_tree_attr_t* next
        myhtml_tree_attr_t* prev

        mycore_string_t key
        mycore_string_t value

        size_t raw_key_begin
        size_t raw_key_length
        size_t raw_value_begin
        size_t raw_value_length

    myhtml_t * myhtml_create()
    mystatus_t myhtml_init(myhtml_t* myhtml, myhtml_options opt, size_t thread_count, size_t queue_size)
    myhtml_tree_t * myhtml_tree_create()
    mystatus_t myhtml_tree_init(myhtml_tree_t* tree, myhtml_t* myhtml)
    mystatus_t myhtml_parse(myhtml_tree_t* tree, myencoding_t encoding, const char* html, size_t html_size)

    myhtml_tree_attr_t* myhtml_node_attribute_first(myhtml_tree_node_t* node)
    myhtml_tree_attr_t* myhtml_attribute_by_key(myhtml_tree_node_t *node, const char *key, size_t key_len)
    const char* myhtml_node_text(myhtml_tree_node_t *node, size_t *length)
    mycore_string_t * myhtml_node_string(myhtml_tree_node_t *node)
    const char * myhtml_tag_name_by_id(myhtml_tree_t* tree, myhtml_tag_id_t tag_id, size_t *length)

    myhtml_collection_t * myhtml_collection_destroy(myhtml_collection_t *collection)
    myhtml_tree_t * myhtml_tree_destroy(myhtml_tree_t* tree)
    myhtml_t* myhtml_destroy(myhtml_t* myhtml)

    myhtml_tree_node_t* myhtml_tree_get_document(myhtml_tree_t* tree)
    myhtml_tree_node_t* myhtml_tree_get_node_body(myhtml_tree_t* tree)
    myhtml_tree_node_t* myhtml_tree_get_node_head(myhtml_tree_t* tree)

    myhtml_collection_t* myhtml_get_nodes_by_name(myhtml_tree_t* tree, myhtml_collection_t *collection,
                                                  const char* name, size_t length, mystatus_t *status)

    void myhtml_node_delete(myhtml_tree_node_t *node)
    void myhtml_node_delete_recursive(myhtml_tree_node_t *node)
    void myhtml_tree_parse_flags_set(myhtml_tree_t* tree, myhtml_tree_parse_flags_t parse_flags)
    myhtml_tree_node_t * myhtml_node_insert_before(myhtml_tree_node_t *target, myhtml_tree_node_t *node)
    myhtml_tree_node_t * myhtml_node_insert_after(myhtml_tree_node_t *target, myhtml_tree_node_t *node)
    myhtml_tree_node_t * myhtml_node_create(myhtml_tree_t* tree, myhtml_tag_id_t tag_id, myhtml_namespace ns)
    myhtml_tree_node_t * myhtml_node_clone_deep(myhtml_tree_t* dest_tree, myhtml_tree_node_t* src)
    myhtml_tree_node_t * myhtml_node_append_child(myhtml_tree_node_t* target, myhtml_tree_node_t* node)

    mycore_string_t * myhtml_node_text_set(myhtml_tree_node_t *node, const char* text, size_t length,
                                           myencoding_t encoding)
    myhtml_tree_attr_t * myhtml_attribute_by_key(myhtml_tree_node_t *node, const char *key, size_t key_len)
    myhtml_tree_attr_t * myhtml_attribute_remove_by_key(myhtml_tree_node_t *node, const char *key, size_t key_len)
    myhtml_tree_attr_t * myhtml_attribute_add(myhtml_tree_node_t *node, const char *key, size_t key_len,
                                              const char *value, size_t value_len, myencoding_t encoding)

    myhtml_tree_node_t * myhtml_node_insert_to_appropriate_place(myhtml_tree_node_t *target, myhtml_tree_node_t *node)

cdef extern from "myhtml/tree.h" nogil:
    myhtml_tree_node_t * myhtml_tree_node_clone(myhtml_tree_node_t* node)
    myhtml_tree_node_t * myhtml_tree_node_insert_root(myhtml_tree_t* tree, myhtml_token_node_t* token,
                                                      myhtml_namespace ns)
    void myhtml_tree_node_add_child(myhtml_tree_node_t* root, myhtml_tree_node_t* node)

cdef extern from "myhtml/serialization.h" nogil:
    mystatus_t myhtml_serialization(myhtml_tree_node_t* scope_node, mycore_string_raw_t* str)


cdef extern from "myencoding/encoding.h" nogil:
    ctypedef enum myencoding_t:
        MyENCODING_DEFAULT        = 0x00
        # MyENCODING_AUTO           = 0x01  // future
        MyENCODING_NOT_DETERMINED = 0x02
        MyENCODING_UTF_8          = 0x00  # default encoding
        MyENCODING_UTF_16LE       = 0x04
        MyENCODING_UTF_16BE       = 0x05
        MyENCODING_X_USER_DEFINED = 0x06
        MyENCODING_BIG5           = 0x07
        MyENCODING_EUC_JP         = 0x08
        MyENCODING_EUC_KR         = 0x09
        MyENCODING_GB18030        = 0x0a
        MyENCODING_GBK            = 0x0b
        MyENCODING_IBM866         = 0x0c
        MyENCODING_ISO_2022_JP    = 0x0d
        MyENCODING_ISO_8859_10    = 0x0e
        MyENCODING_ISO_8859_13    = 0x0f
        MyENCODING_ISO_8859_14    = 0x10
        MyENCODING_ISO_8859_15    = 0x11
        MyENCODING_ISO_8859_16    = 0x12
        MyENCODING_ISO_8859_2     = 0x13
        MyENCODING_ISO_8859_3     = 0x14
        MyENCODING_ISO_8859_4     = 0x15
        MyENCODING_ISO_8859_5     = 0x16
        MyENCODING_ISO_8859_6     = 0x17
        MyENCODING_ISO_8859_7     = 0x18
        MyENCODING_ISO_8859_8     = 0x19
        MyENCODING_ISO_8859_8_I   = 0x1a
        MyENCODING_KOI8_R         = 0x1b
        MyENCODING_KOI8_U         = 0x1c
        MyENCODING_MACINTOSH      = 0x1d
        MyENCODING_SHIFT_JIS      = 0x1e
        MyENCODING_WINDOWS_1250   = 0x1f
        MyENCODING_WINDOWS_1251   = 0x20
        MyENCODING_WINDOWS_1252   = 0x21
        MyENCODING_WINDOWS_1253   = 0x22
        MyENCODING_WINDOWS_1254   = 0x23
        MyENCODING_WINDOWS_1255   = 0x24
        MyENCODING_WINDOWS_1256   = 0x25
        MyENCODING_WINDOWS_1257   = 0x26
        MyENCODING_WINDOWS_1258   = 0x27
        MyENCODING_WINDOWS_874    = 0x28
        MyENCODING_X_MAC_CYRILLIC = 0x29
        MyENCODING_LAST_ENTRY     = 0x2a

    bint myencoding_detect_bom(const char *text, size_t length, myencoding_t *encoding)
    bint myencoding_detect(const char *text, size_t length, myencoding_t *encoding)
    myencoding_t myencoding_prescan_stream_to_determine_encoding(const char *data, size_t data_size)
    const char* myencoding_name_by_id(myencoding_t encoding, size_t *length)


cdef extern from "mycss/mycss.h" nogil:
    ctypedef struct mycss_entry_t:
        # not completed struct
        mycss_t* mycss

    ctypedef struct mycss_t

    ctypedef struct mycss_selectors_t

    ctypedef struct mycss_selectors_entries_list_t
    ctypedef struct mycss_declaration_entry_t

    ctypedef enum mycss_selectors_flags:
        MyCSS_SELECTORS_FLAGS_UNDEF         = 0x00
        MyCSS_SELECTORS_FLAGS_SELECTOR_BAD  = 0x01
    ctypedef mycss_selectors_flags mycss_selectors_flags_t

    ctypedef struct mycss_selectors_list_t:
        mycss_selectors_entries_list_t* entries_list
        size_t entries_list_length

        mycss_declaration_entry_t* declaration_entry

        mycss_selectors_flags_t flags

        mycss_selectors_list_t* parent
        mycss_selectors_list_t* next
        mycss_selectors_list_t* prev

    # CSS init routines
    mycss_t * mycss_create()
    mystatus_t mycss_init(mycss_t* mycss)
    mycss_entry_t * mycss_entry_create()
    mystatus_t mycss_entry_init(mycss_t* mycss, mycss_entry_t* entry)

    mycss_selectors_list_t * mycss_selectors_parse(mycss_selectors_t* selectors, myencoding_t encoding,
                                                   const char* data, size_t data_size, mystatus_t* out_status)
    mycss_selectors_t * mycss_entry_selectors(mycss_entry_t* entry)

    mycss_selectors_list_t * mycss_selectors_list_destroy(mycss_selectors_t* selectors,
                                                          mycss_selectors_list_t* selectors_list, bint self_destroy)
    mycss_entry_t * mycss_entry_destroy(mycss_entry_t* entry, bint self_destroy)
    mycss_t * mycss_destroy(mycss_t* mycss, bint self_destroy)


cdef extern from "modest/finder/finder.h" nogil:
    ctypedef struct modest_finder_t
    modest_finder_t* modest_finder_create_simple()
    mystatus_t modest_finder_by_selectors_list(modest_finder_t* finder, myhtml_tree_node_t* scope_node,
                                               mycss_selectors_list_t* selector_list, myhtml_collection_t** collection)
    modest_finder_t * modest_finder_destroy(modest_finder_t* finder, bint self_destroy)


cdef class HTMLParser:
    cdef myhtml_tree_t *html_tree
    cdef public bint detect_encoding
    cdef public bint use_meta_tags
    cdef myencoding_t _encoding
    cdef public unicode decode_errors
    cdef public bytes raw_html
    cdef object cached_script_texts
    cdef object cached_script_srcs

    cdef void _detect_encoding(self, char* html, size_t html_len) nogil
    cdef int _parse_html(self, char* html, size_t html_len) except -1

    @staticmethod
    cdef HTMLParser from_tree(
        myhtml_tree_t * tree, bytes raw_html, bint detect_encoding, bint use_meta_tags, str decode_errors,
        myencoding_t encoding
    )


cdef class Stack:
    cdef size_t capacity
    cdef size_t top
    cdef myhtml_tree_node_t ** _stack

    cdef bint is_empty(self)
    cdef int push(self, myhtml_tree_node_t* res) except -1
    cdef myhtml_tree_node_t * pop(self)
    cdef int resize(self) except -1



cimport cython
from cpython.exc cimport PyErr_SetObject


@cython.final
cdef class CSSSelector:

    cdef char *c_selector
    cdef mycss_entry_t *css_entry
    cdef modest_finder_t *finder
    cdef mycss_selectors_list_t *selectors_list

    def __init__(self, str selector):

        selector_pybyte = selector.encode('UTF-8')
        self.c_selector = selector_pybyte

        # In order to propagate errors these methods should return no value
        self._create_css_parser()
        self._prepare_selector(self.css_entry, self.c_selector, len(self.c_selector))
        self.finder = modest_finder_create_simple()

    cdef myhtml_collection_t* find(self, myhtml_tree_node_t* scope):
        """Find all possible matches."""

        cdef myhtml_collection_t *collection

        collection = NULL
        modest_finder_by_selectors_list(self.finder, scope, self.selectors_list, &collection)

        return collection

    cdef int _create_css_parser(self) except -1:
        cdef mystatus_t status

        cdef mycss_t *mycss = mycss_create()
        status = mycss_init(mycss)

        if status != 0:
            PyErr_SetObject(RuntimeError, "Can't init MyCSS object.")
            return -1

        self.css_entry = mycss_entry_create()
        status = mycss_entry_init(mycss, self.css_entry)

        if status != 0:
            PyErr_SetObject(RuntimeError, "Can't init MyCSS Entry object.")
            return -1
        return 0

    cdef int _prepare_selector(self, mycss_entry_t *css_entry, const char *selector, size_t selector_size) except -1:
        cdef mystatus_t out_status
        self.selectors_list = mycss_selectors_parse(mycss_entry_selectors(css_entry), myencoding_t.MyENCODING_UTF_8,
                                                    selector, selector_size, &out_status)

        if (self.selectors_list == NULL) or (self.selectors_list.flags and MyCSS_SELECTORS_FLAGS_SELECTOR_BAD):
            PyErr_SetObject(ValueError, "Bad CSS Selectors: %s" % self.c_selector.decode('utf-8'))
            return -1
        return 0

    def __dealloc__(self):
        mycss_selectors_list_destroy(mycss_entry_selectors(self.css_entry), self.selectors_list, 1)
        modest_finder_destroy(self.finder, 1)

        cdef mycss_t *mycss = self.css_entry.mycss
        mycss_entry_destroy(self.css_entry, 1)
        mycss_destroy(mycss, 1)


cdef class Selector:
    """An advanced CSS selector that supports additional operations.

    Think of it as a toolkit that mimics some of the features of XPath.

    Please note, this is an experimental feature that can change in the future.
    """
    cdef Node node
    cdef list nodes

    def __init__(self, Node node, str query):
        """custom init, because __cinit__ doesn't accept C types"""
        self.node = node
        self.nodes = find_nodes(node.parser, node.node, query) if query else [node, ]

    cpdef css(self, str query):
        """Evaluate CSS selector against current scope."""
        cdef Node current_node
        nodes = list()
        for node in self.nodes:
            current_node = node
            nodes.extend(find_nodes(self.node.parser, current_node.node, query))
        self.nodes = nodes
        return self

    @property
    def matches(self):
        """Returns all possible matches"""
        return self.nodes

    @property
    def any_matches(self):
        """Returns True if there are any matches"""
        return bool(self.nodes)

    def text_contains(self, str text, bool deep=True, str separator='', bool strip=False):
        """Filter all current matches given text."""
        nodes = []
        cdef Node node
        for node in self.nodes:
            node_text = node.text(deep=deep, separator=separator, strip=strip)
            if node_text and text in node_text:
                nodes.append(node)
        self.nodes = nodes
        return self

    def any_text_contains(self, str text, bool deep=True, str separator='', bool strip=False):
        """Returns True if any node in the current search scope contains specified text"""
        nodes = []
        cdef Node node
        for node in self.nodes:
            node_text = node.text(deep=deep, separator=separator, strip=strip)
            if node_text and text in node_text:
                return True
        return False

    def attribute_longer_than(self, str attribute, int length, str start  = None):
        """Filter all current matches by attribute length.

        Similar to `string-length` in XPath.
        """
        nodes = []
        for node in self.nodes:
            attr = node.attributes.get(attribute)
            if attr and start and start in attr:
                attr = attr[attr.find(start) + len(start):]
            if len(attr) > length:
                nodes.append(node)
        self.nodes = nodes
        return self

    def any_attribute_longer_than(self, str attribute, int length, str start  = None):
        """Returns True any href attribute longer than a specified length.

        Similar to `string-length` in XPath.
        """
        cdef list nodes = []
        cdef Node node
        for node in self.nodes:
            attr = node.attributes.get(attribute)
            if attr and start and start in attr:
                attr = attr[attr.find(start) + len(start):]
            if len(attr) > length:
                return True
        return False

    def __bool__(self):
        return bool(self.nodes)

cdef find_nodes(HTMLParser parser, myhtml_tree_node_t *node, str query):
    cdef myhtml_collection_t *collection
    cdef CSSSelector selector = CSSSelector(query)
    cdef Node n
    cdef list result = []
    collection = selector.find(node)

    if collection == NULL:
        return result

    for i in range(collection.length):
        n = Node.new(collection.list[i], parser)
        result.append(n)
    myhtml_collection_destroy(collection)
    return result


cdef bool find_matches(HTMLParser parser, myhtml_tree_node_t *node, tuple selectors):
    cdef myhtml_collection_t *collection
    cdef CSSSelector selector
    cdef int collection_size
    cdef str query

    for query in selectors:
        selector = CSSSelector(query)
        collection_size = 0
        collection = NULL

        collection = selector.find(node)
        if collection == NULL:
            continue

        collection_size = collection.length
        myhtml_collection_destroy(collection)
        if collection_size > 0:
            return True
    return False
